#include "common.h"
#include "reception.h"

void update_recv_window(byte type,byte level, int data_size,ADOC_session *ADOC_global){
    /* Implementation ensures that if you are here, there is at least 
    enough space in the window to store one packet*/

    ADOC_global->header_window[ADOC_global->header_w].type=type;
    ADOC_global->header_window[ADOC_global->header_w].level=level;
    ADOC_global->header_window[ADOC_global->header_w].data_size=data_size;
    ADOC_global->header_window[ADOC_global->header_w].data=ADOC_global->data_w_pointer;
    /* Tell the other thread that one packet has been added to the window
        but not for the first packet because data arrive with the next packet*/
    if(ADOC_global->ADOC_flag)
        adoc_sem_post(&(ADOC_global->r_sem));
    else
        ADOC_global->ADOC_flag=1;

    /*Wait if the window is full now*/
    adoc_sem_wait(&(ADOC_global->w_sem));

    /*Updating the wrinting pointer of the data window cyclically*/
    ADOC_global->data_w_pointer+=ADOC_PACKET_SIZE;
    /* If dealing with an uncompressed packet or the last packet of a compressed 
        buffer or etc... */
    if((type!=ADOC_TYPE_COMP) && (ADOC_global->data_w_pointer>ADOC_global->data_window+(ADOC_MAX_RECV_PACKETS)*ADOC_PACKET_SIZE))
        ADOC_global->data_w_pointer=ADOC_global->data_window;

    /*Updating the writing pointer of the packet window cyclically */
    ADOC_global->header_w++;
    if(ADOC_global->header_w>=ADOC_MAX_RECV_PACKETS)
        ADOC_global->header_w=0;

}


ssize_t recv_iovec(int socket,struct iovec *iov,int iovec_size,int header_size,
                   size_t data_size,ADOC_session *ADOC_global){

    ssize_t ret,nb_recv,res,sum_len;

    int i;

    for(;;) {
        ret=readv(socket,iov,iovec_size);

        if(ret==0) /*socket close by the peer*/
            return 0;

        if(ret<0){
            if(errno == EINTR)
            {
                /*fprintf(stderr,".");*/
                continue;
            }

            if (errno==ECONNRESET)/*connection reset by peer*/
                return 0;
            else{
                fprintf(stderr,"AdOC : readv error (%d) while reading socket : %s\n",errno,strerror(errno));
                return ADOC_ERROR_WITH_ERRNO_SET;
            }
        }

        break;
    }

    res=ret;
    if(res==data_size+header_size){
        return res;
    }


    if(res>header_size){
        res-=header_size;
        nb_recv=READ(socket,(byte *)iov[3].iov_base+res,iov[3].iov_len-res);
        if(nb_recv<0)
            return ADOC_ERROR_WITH_ERRNO_SET;
        else
            return res+nb_recv+header_size;
    }else{
        sum_len=0;
        for(i=0;i<iovec_size;i++){
            if(res<sum_len+iov[i].iov_len){
                nb_recv=READ(socket,(byte *)iov[i].iov_base+res-sum_len,
                             iov[i].iov_len-(res-sum_len));
                if(nb_recv<0)
                    return ADOC_ERROR_WITH_ERRNO_SET;
                res+=nb_recv;
            }
            sum_len+=iov[i].iov_len;
        }
        return res;
    }

}


/*This thread has in charge of reading packets from the 
network and store them into the cicular window*/ 
void * reception_thread(void *arg){ 
    int socket;
    ssize_t res;    
    uint8_t type_t,level_t;
    byte type,level=0;
    uint32_t data_size_t;
    size_t data_size=0;
    struct iovec iov[4];
    int header_size;
    ADOC_session  *ADOC_global=(ADOC_session*)arg;

    socket=ADOC_global->socket;
    /*header_size is the size of first 3 vector of iov.
        It must have a constant value and must not depend
        on the architecture 32 bits or 64 bits or whatever*/
    data_size_t=htonl(0);
    header_size=1+1+4;
    iov[0].iov_base=(void*)&type_t;
    iov[0].iov_len=1;
    iov[1].iov_base=(void*)&level_t;
    iov[1].iov_len=1;
    iov[2].iov_base=(void*)&data_size_t;
    iov[2].iov_len=4;
    iov[3].iov_base=(void*)NULL;
    iov[3].iov_len=0;
    do{
        res=recv_iovec(socket, iov, 4, header_size, data_size, ADOC_global);
        type=(byte)(type_t);
        level=(byte)(level_t);
        /*fprintf(stderr,"res=%ld\n",(long)res);*/
        /*fprintf(stderr,"type:%d, level:%d\n",type,level);*/
        if(res==0){/* Attempt to read a socket cleanly closed by peer*/
            update_recv_window(ADOC_TYPE_BS,0,0,ADOC_global);
            update_recv_window(ADOC_TYPE_SESSION_SIZE,0,0,ADOC_global);
            update_recv_window(ADOC_TYPE_EOF,0,0,ADOC_global);
            update_recv_window(ADOC_TYPE_EOF,0,0,ADOC_global);
            pthread_exit(0);

        }
        else if(res<0){
            ADOC_global->error=res;
            /*unblock other thread */
            adoc_sem_post(&(ADOC_global->r_sem));
            pthread_exit(0);
        } else{
            data_size=(size_t)ntohl(data_size_t);
            /*fprintf(stderr," ## data_size = %ld\n", data_size);*/
            switch(type){
                case ADOC_TYPE_LAST_PACKET:
                case ADOC_TYPE_COMP:
                case ADOC_TYPE_UNCOMP:
                    iov[3].iov_len=data_size;
                    /* Data arrive with the next packet */
                    iov[3].iov_base=(void*)ADOC_global->data_w_pointer;
                    break;
                case ADOC_TYPE_BS:
                case ADOC_TYPE_SESSION_SIZE:
                case ADOC_TYPE_EOF:
                    iov[3].iov_len=0;
                    break;
                default:
                    fprintf(stderr,"@#$#AdOC : Unknown packet type(%d), reception\n",type);
                    ADOC_global->error=ADOC_UNKNOWN_PACKET;
                    pthread_exit(0);
            }
        }
        update_recv_window(type,level,data_size,ADOC_global);

        /*Check whenever an error has been generated by 
            the decompression thread*/
        if(ADOC_global->error!=0){
            pthread_exit(0);
        }
    }while(type!=ADOC_TYPE_EOF);

update_recv_window(ADOC_TYPE_EOF,0,0,ADOC_global);
pthread_exit(0);
/*Make compiler happy*/
return NULL; 
}


